from abc import ABC, abstractmethod
from typing import List
from rltk.tokenizer.crf_tokenizer import crf_tokenizer as dig_tokenizer


class Tokenizer(ABC):
    """
    Abstract tokenizer
    """

    @abstractmethod
    def tokenize(self, s: str) -> List[str]:
        """
        Apply tokenizer

        Args:
            s (str): String to tokenize.

        Returns:
            List[str]: Tokenized list. It won't do token deduplication.
        """
        raise NotImplementedError


class CRFTokenizer(Tokenizer):
    """
    CRFTokenizer: this uses old DIG CRFTokenizer
    """

    def __init__(self, *args, **kwargs) -> None:
        self._t = dig_tokenizer.CrfTokenizer(*args, **kwargs)

    def tokenize(self, s: str) -> List[str]:
        return self._t.tokenize(s)


class WordTokenizer(Tokenizer):
    """
    Word Tokenizer: tokenize word by white space

    Args:
        remove_empty (bool, optional): If set, empty token will be removed. Defaults to False.
    """

    def __init__(self, remove_empty: bool = False) -> None:
        self._remove_empty = remove_empty

    def tokenize(self, s: str) -> List[str]:
        s = s.split(' ')
        if self._remove_empty:
            return list(filter(lambda x: len(x) != 0, s))
        else:
            return s


class NGramTokenizer(Tokenizer):
    """
    NGrame Tokenizer

    Args:
         n (int): n.
         place_holder (str, optional): String to fill pad and separator. Defaults to white space (' ').
         padded (bool, optional): If set, head the tail will be filled with pad. Defaults to False.
    """

    def __init__(self, n: int, place_holder: str = ' ', padded: bool = False,
                 base_tokenizer: Tokenizer = None) -> None:
        self._n = n
        self._place_holder = place_holder
        self._padded = padded
        self._base_tokenizer = base_tokenizer if base_tokenizer else WordTokenizer()

    def tokenize(self, s: str) -> List[str]:
        if len(s) == 0:
            return []
        if self._padded:
            pad = self._place_holder * (self._n - 1)
            s = pad + s + pad
        s = self._base_tokenizer.tokenize(s)
        s = self._place_holder.join(s)
        if len(s) < self._n:
            return [s]
        return [s[i:i + self._n] for i in range(len(s) - self._n + 1)]



import unicodedata
import warnings

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from rltk.dataset import Dataset
    from rltk.blocking.block import Block
    from rltk.evaluation.ground_truth import GroundTruth


def check_for_none(*args):
    for arg in args:
        if arg is None:
            raise ValueError('Missing parameter')


def check_for_type(type, *args):
    for arg in args:
        if not isinstance(arg, type):
            raise TypeError('Wrong type of parameter')


def unicode_normalize(s):
    return unicodedata.normalize('NFKD', s)


def convert_list_to_set(s):
    if isinstance(s, list):
        s = set(s)
    return s


def candidate_pairs(dataset1: 'Dataset',
                     dataset2: 'Dataset' = None,
                     block: 'Block' = None,
                     ground_truth: 'GroundTruth' = None):
    """
    Generate candidate pairs to compare.

    Args:
        dataset1 (Dataset): dataset 1.
        dataset2 (Dataset, optional): dataset 2. If it's not provided, it will be a de-duplication task.
        block (Block, optional): Block.
        ground_truth (GroundTruth, optional): Ground truth.
    """
    if block and not ground_truth:
        if not dataset2:
            for _, id1, id2 in block.pairwise(dataset1.id):
                yield dataset1.get_record(id1), dataset1.get_record(id2)
        else:
            for _, id1, id2 in block.pairwise(dataset1.id, dataset2.id):
                yield dataset1.get_record(id1), dataset2.get_record(id2)
    elif ground_truth and not block:
        if not dataset2:
            for id1, id2, label in ground_truth:
                yield dataset1.get_record(id1), dataset1.get_record(id2)
        else:
            for id1, id2, label in ground_truth:
                yield dataset1.get_record(id1), dataset2.get_record(id2)
    elif ground_truth and block:
        if not dataset2:
            for _, id1, id2 in block.pairwise(dataset1.id):
                if ground_truth.is_member(id1, id2):
                    yield dataset1.get_record(id1), dataset1.get_record(id2)
        else:
            for _, id1, id2 in block.pairwise(dataset1.id, dataset2.id):
                if ground_truth.is_member(id1, id2):
                    yield dataset1.get_record(id1), dataset2.get_record(id2)
    else:
        if not dataset2:
            skip_offset = 0
            for r1 in dataset1:
                for offset, r2 in enumerate(dataset1):
                    if offset < skip_offset:
                        continue
                    if r1.id == r2.id:
                        continue
                    yield r1, r2
                skip_offset += 1
        else:
            for r1 in dataset1:
                for r2 in dataset2:
                    yield r1, r2


get_record_pairs = candidate_pairs


class ModuleImportWarning(UserWarning):
    pass


def module_importer(module_names: str, dependencies: str, notes: str = None):
    if isinstance(dependencies, str):
        dependencies = [dependencies]

    def module():
        try:
            return __import__(module_names)
        except ImportError:
            warning_msg = '\n-----------------------------------\n'
            warning_msg += '\nImport Dependencies Error\n'

            if len(dependencies) > 0:
                warning_msg += '\nPlease install dependencies:\n'
                for d in dependencies:
                    warning_msg += d + '\n'

            if notes:
                warning_msg += notes

            warning_msg += '\n-----------------------------------'
            warnings.warn(warning_msg, ModuleImportWarning)
            exit(500)

    return module



from typing import Callable
from functools import cmp_to_key

from rltk.blocking.block_generator import BlockGenerator
from rltk.blocking.block import Block
from rltk.blocking.block_black_list import BlockBlackList


class SortedNeighbourhoodBlockGenerator(BlockGenerator):
    """
    Sorted Neighbourhood Blocker.
    
    Args:
        window_size (int, optional): Window size. Defaults to 3.
        comparator (Callable, optional): Define how to compare two tokens t1 and t2.
                            The signature is `comparator(t1: str, t2: str) -> int`.
                            If return is 0, t1 equals t2; if return is -1, t1 is less than t2;
                            if return is 1, t1 is greater than t2.
                            Defaults to None, which uses Python's default string comparison.
        block_id_prefix (str, optional): The block id prefix of each block.
                                        Defaults to "sorted_neighbourhood_".
    """
    def __init__(self, window_size: int = 3, comparator: Callable = None, block_id_prefix='sorted_neighbourhood_'):
        if comparator is None:
            comparator = self._default_comparator
        self.window_size = window_size
        self.comparator = comparator
        self.block_id_prefix = block_id_prefix

    def block(self, dataset, function_: Callable = None, property_: str = None,
              block: Block = None, block_black_list: BlockBlackList = None, base_on: Block = None):
        """
        The return of `property_` or `function_` should be a vector (list).
        """
        block = super()._block_args_check(function_, property_, block)

        if base_on:
            for block_id, dataset_id, record_id in base_on:
                if dataset.id == dataset_id:
                    r = dataset.get_record(record_id)
                    value = function_(r) if function_ else getattr(r, property_)
                    if not isinstance(value, (list, set)):
                        value = value(set)
                    for v in value:
                        if not isinstance(v, str):
                            raise ValueError('Elements in return list should be string')
                        if block_black_list and block_black_list.has(v):
                            continue
                        v = block_id + '-' + v
                        block.add(v, dataset.id, r.id)
                        if block_black_list:
                            block_black_list.add(v, block)

        else:
            for r in dataset:
                value = function_(r) if function_ else getattr(r, property_)
                if not isinstance(value, (list, set)):
                    value = set(value)
                for v in value:
                    if not isinstance(v, str):
                        raise ValueError('Elements in return list should be string')
                    if block_black_list and block_black_list.has(v):
                        continue
                    block.add(v, dataset.id, r.id)
                    if block_black_list:
                        block_black_list.add(v, block)

        return block

    def generate(self, block1: Block, block2: Block, output_block: Block = None):
        output_block = BlockGenerator._generate_args_check(output_block)

        # TODO: in-memory operations here, need to update
        # concatenation
        all_records = []
        for block_id, ds_id, record_id in block1:
            all_records.append((block_id, ds_id, record_id))
        for block_id, ds_id, record_id in block2:
            all_records.append((block_id, ds_id, record_id))
        sorted_all_records = sorted(all_records, key=cmp_to_key(self._comparator_wrapper))

        # apply slide window
        for i in range(len(sorted_all_records) - self.window_size + 1):
            block_id = self.block_id_prefix + str(i)
            for j in range(self.window_size):
                record = sorted_all_records[i + j]
                output_block.add(block_id, record[1], record[2])

        return output_block

    def _comparator_wrapper(self, t1, t2):
        return self.comparator(t1[0], t2[0])

    @staticmethod
    def _default_comparator(t1, t2):
        return 0 if t1 == t2 else (1 if t1 > t2 else -1)



__version__ = '2.0.0-a017'

from rltk.record import Record, AutoGeneratedRecord,\
    cached_property, generate_record_property_cache, validate_record, remove_raw_object, set_id
from rltk.dataset import Dataset
from rltk.io import *
from rltk.similarity import *
from rltk.blocking import *
from rltk.tokenizer import *
from rltk.evaluation import *
from rltk.utils import candidate_pairs, get_record_pairs
import rltk.cli
import rltk.remote
